from requests.exceptions import HTTPError

from netbox_netprod_importer.exceptions import GenericNetboxError


def is_macaddr(macaddr):
    macaddr_simplified = macaddr.replace(":", "").replace(".", "")
    if len(macaddr_simplified) != 12:
        return False

    try:
        hex(int(macaddr_simplified, 16))
    except ValueError:
        return False

    return True


def macaddr_to_int(macaddr):
    if not macaddr:
        return 0
    macaddr_simplified = macaddr.replace(":", "").replace(".", "")
    return int(macaddr_simplified, 16)


def generic_netbox_error(func):
    """
    Convert an HTTP error to a more explicit exception
    """
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HTTPError as e:
            raise GenericNetboxError(e)
    return wrapper



from enum import Enum


class NetboxInterfaceTypes(Enum):
    cfp = "CFP (100GE)"
    cfp2 = "CFP2 (100GE)"
    eth100 = "100BASE-TX (10/100ME)"
    eth1000 = "1000BASE-T (1GE)"
    sfp = "SFP (1GE)"
    sfp_plus = "SFP+ (10GE)"
    sfp28 = "SFP28 (25GE)"
    qsfp_plus = "QSFP+ (40GE)"
    qsfp28 = "QSFP+ (40GE)"
    xenpack = "XENPACK (10GE)"
    x2 = "X2"
    xfp = "XFP (10GE)"



from simplejson.errors import JSONDecodeError

class NoReverseFoundError(Exception):
    def __init__(self, host):
        super().__init__("No reverse found for host {}".format(host))


class TypeCouldNotBeParsedError(Exception):
    pass


class DeviceNotFoundError(Exception):
    def __init__(self, hostname):
        super().__init__("Device {} not found on Netbox".format(hostname))
        self.hostname = hostname


class _NetboxPushingError(Exception):
    netbox_exc = None

    def _extract_netbox_error(self):
        try:
            return self.netbox_exc.response.json()
        except JSONDecodeError:
            return self.netbox_exc.response.content.decode()


class GenericNetboxError(_NetboxPushingError):
    def __init__(self, netbox_req_exc):
        self.netbox_exc = netbox_req_exc

        reason = self._extract_netbox_error()
        super().__init__("{} -- {}".format(netbox_req_exc, reason))


class IPPushingError(_NetboxPushingError):
    def __init__(self, ip, netbox_req_exc):
        self.ip = ip
        self.netbox_exc = netbox_req_exc

        reason = self._extract_netbox_error()
        super().__init__("IP {} could not be created: {} -- {}".format(
            ip, netbox_req_exc, reason
        ))


class NetIfPushingError(_NetboxPushingError):
    def __init__(self, netif_name, netbox_req_exc):
        self.netif_name = netif_name
        self.netbox_exc = netbox_req_exc

        reason = self._extract_netbox_error()
        super().__init__("Interface {} could not be created: {} -- {}".format(
            netif_name, netbox_req_exc, reason
        ))


class NetInterfaceNotFoundError(Exception):
    def __init__(self, netif):
        super().__init__(
            "Network interface {} not found on Netbox".format(netif)
        )
        self.netif = netif


class DeviceNotSupportedError(Exception):
    def __init__(self, hostname):
        super().__init__("Device {} not supported".format(hostname))
        self.hostname = hostname



import re
import logging
from collections import defaultdict

from netbox_netprod_importer.exceptions import TypeCouldNotBeParsedError
from netbox_netprod_importer.vendors.constants import NetboxInterfaceTypes
from .constants import InterfacesRegex
from .base import CiscoParser


logger = logging.getLogger("netbox_importer")


class IOSParser(CiscoParser):
    def get_interfaces_lag(self, interfaces):
        super().get_interfaces_lag(interfaces)

        interfaces_lag = defaultdict(list)
        for interface in sorted(interfaces):
            cmd = "show run interface {}".format(interface)
            interface_conf_dump = self.device.cli([cmd])[cmd]

            channel_group_match = re.search(
                r"^\s*channel-group (\S*)", interface_conf_dump, re.MULTILINE
            )
            if channel_group_match:
                port_channel_id = channel_group_match.groups()[0]
            else:
                continue

            interfaces_lag[interface] = "port-channel{}".format(
                port_channel_id
            )

        return interfaces_lag

    def get_interface_type(self, interface):
        super().get_interface_type(interface)
        if re.search(r"^Vlan(\d*)|^Tunnel(\d+)", interface):
            return "Virtual"

        interface_type = "Other"
        try:
            cisco_if_type = self._guess_type_from_if_type(interface)

            for pattern_iftype in InterfacesRegex:
                pattern = pattern_iftype.value
                if re.match(pattern, cisco_if_type):
                    return getattr(
                        NetboxInterfaceTypes, pattern_iftype.name
                    ).value
        except TypeCouldNotBeParsedError:
            pass

        return interface_type

    def _guess_type_from_if_type(self, interface):
        from pynxos.errors import CLIError

        try:
            abrev_if = self.get_abrev_if(interface)
            return self._get_ifstatus_by_abrev_if()[abrev_if]
        except (KeyError, CLIError):
            raise TypeCouldNotBeParsedError()

    def _get_ifstatus_by_abrev_if(self):
        cmd = "show interface status"

        if not self.cache.get("ifstatus"):
            status_conf_dump = self.device.cli([cmd])[cmd].strip()
            self.cache["ifstatus"] = {}
            start = status_conf_dump.find('Type')
            for l in status_conf_dump.splitlines()[1:]:
                split_l = l.split(maxsplit=1)
                if_abrev = split_l[0]
                try:
                    if_type = l[start:]
                except:
                    if_type = None

                self.cache["ifstatus"][if_abrev] = if_type

        return self.cache["ifstatus"]

    def get_detailed_cdp_neighbours(self):
        cmd = "show cdp neighbors detail"
        cmd_output = self.device.cli([cmd])[cmd]
        cmd_output = re.split('---+\n', cmd_output)
        neighbours = []
        for cdp_port_info in cmd_output:
            if len(cdp_port_info):
                info_port = re.split('\n\n', cdp_port_info)[0].split('\n')
                hostname = local_port = port = ''
                for line in info_port:
                    if line.startswith('Device'):
                        hostname = line.split(':')[1].strip()
                    elif line.startswith('Interface'):
                        local_port, port = line.split(',')
                        local_port = local_port.split(':')[1].strip()
                        port = port.split(':')[1].strip()
                neighbours.append(
                    {
                        "local_port": local_port,
                        "hostname": hostname,
                        "port": port
                    }
                )
        for n in neighbours:
            yield n

    def get_interface_mode(self, interface):
        from pynxos.errors import CLIError
        try:
            return self._get_interfaces_mode()[
                self.get_abrev_if(interface)].get("oper_mode")
        except (KeyError, CLIError):
            logger.debug("Switch %s, show interface switchport cmd error",
                         self.device.hostname)
        return None

    def get_interface_access_vlan(self, interface):
        from pynxos.errors import CLIError
        try:
            return self._get_interfaces_mode()[
                self.get_abrev_if(interface)].get("access_vlan")
        except (KeyError, CLIError):
            logger.debug("Switch %s, show interface switchport cmd error",
                         self.device.hostname)
        return None

    def get_interface_netive_vlan(self, interface):
        from pynxos.errors import CLIError
        try:
            return self._get_interfaces_mode()[
                self.get_abrev_if(interface)].get("native_vlan")
        except (KeyError, CLIError):
            logger.debug("Switch %s, show interface switchport cmd error",
                         self.device.hostname)
        return None

    def _get_interfaces_mode(self):
        cmd = "show interface switchport"

        if not self.cache.get("mode"):
            mode_conf_dump = self.device.cli([cmd])[cmd]
            mode_conf_lines = re.split(r"(^Name: \S+$)", mode_conf_dump, flags=re.M)
            mode_conf_lines.pop(0)
            if len(mode_conf_lines) % 2 != 0:
                raise ValueError("Unexpected output data in '{}':\n\n{}".format(
                    cmd, mode_conf_lines
                ))
            mode_conf_iter = iter(mode_conf_lines)
            try:
                new_mode = [line + next(mode_conf_iter, "") for line in mode_conf_iter]
            except TypeError:
                raise ValueError()

            interfaces_mode = {}
            for entry in new_mode:
                grp = [
                    r"^Name:\s+(?P<interface>\S+)",
                    r"^Administrative Mode:\s+(?P<oper_mode>static access|trunk|access)",
                    r"^Access Mode VLAN:\s+(?P<access_vlan>\d+)",
                    r"^Trunking Native Mode VLAN:\s+(?P<native_valn>\d+)"
                ]
                inf_mode = {}
                for g in grp:
                    find = re.search(g, entry, re.MULTILINE)
                    if find:
                        inf_mode[find.lastgroup] = find.group(find.lastgroup)
                if inf_mode.get("interface"):
                    interfaces_mode[inf_mode["interface"]] = inf_mode

            self.cache["mode"] = interfaces_mode

        return self.cache["mode"]

    def get_vlans(self):
        """
        Napalm does not support vlan
        but there are issues https://github.com/napalm-automation/napalm/issues/927
        PR: https://github.com/napalm-automation/napalm/pull/948

        :return vlans:
            vlan_id, {
                "name": vlan name,
                "interfaces": list interfaces dict
            }
        """
        interfaces = self.device.get_interfaces()
        interface_dict = {}
        for interface, data in interfaces.items():
            interface_dict[self.get_abrev_if(interface)] = interface

        command = "show vlan all-ports"
        output = self.device.cli([command])[command]
        if output.find("Invalid input detected") >= 0:
            yield from self._get_vlan_from_id(interface_dict)
        else:
            yield from self._get_vlan_all_ports(interface_dict, output)

    def _get_vlan_all_ports(self, interface_dict, output):
        find_regexp = r"^(\d+)\s+(\S+)\s+\S+\s+([A-Z][a-z].*)$"
        find = re.findall(find_regexp, output, re.MULTILINE)
        for v in find:
            yield v[0], {
                "name": v[1],
                "interfaces": [
                    interface_dict[x.strip()] for x in v[2].split(",")
                ],
            }
        find_regexp = r"^(\d+)\s+(\S+)\s+\S+$"
        find = re.findall(find_regexp, output, re.MULTILINE)
        for v in find:
            yield v[0], {"name": v[1], "interfaces": []}

    def _get_vlan_from_id(self, interface_dict):
        command = "show vlan brief"
        output = self.device.cli([command])[command]
        vlan_regexp = r"^(\d+)\s+(\S+)\s+\S+.*$"
        find_vlan = re.findall(vlan_regexp, output, re.MULTILINE)
        for vlan in find_vlan:
            command = "show vlan id {}".format(vlan[0])
            output = self.device.cli([command])[command]
            interface_regex = r"{}\s+{}\s+\S+\s+([A-Z][a-z].*)$".format(
                vlan[0], vlan[1]
            )
            interfaces = re.findall(interface_regex, output, re.MULTILINE)
            if len(interfaces) == 1:
                yield vlan[0], {
                    "name": vlan[1],
                    "interfaces": [
                        interface_dict[x.strip()] for x in interfaces[0].split(",")
                    ],
                }
            elif len(interfaces) == 0:
                yield vlan[0], {"name": vlan[1], "interfaces": []}
            else:
                logger.error(
                    "Switch %s Error parsing for vlan id %s, "
                    "found more values than can be.",
                    self.device.hostname, vlan[0]
                )
                yield None, None



import json
import re
import logging

from netbox_netprod_importer.exceptions import TypeCouldNotBeParsedError
from netbox_netprod_importer.vendors.constants import NetboxInterfaceTypes
from .constants import InterfacesRegex
from .base import CiscoParser
from napalm.nxos.nxos import NXOSDriver
from collections import defaultdict

logger = logging.getLogger("netbox_importer")


class NXOSParser(CiscoParser):
    def get_interface_type(self, interface):
        super().get_interface_type(interface)
        if re.search(r"^Vlan(\d*)|^Tunnel(\d+)", interface):
            return "Virtual"

        fn_order = (
            self._parse_type_from_transceiver,
            self._guess_type_from_if_type
        )

        interface_type = "Other"
        for f in fn_order:
            try:
                return f(interface)
            except TypeCouldNotBeParsedError:
                continue

        return interface_type

    def _parse_type_from_transceiver(self, interface):
        from pynxos.errors import CLIError

        try:
            part_num = self._get_transceiver_by_if()[interface]["partnum"]
        except (json.JSONDecodeError, KeyError, CLIError):
            logger.debug("%s has no transceiver detail", interface)
            raise TypeCouldNotBeParsedError()

        for pattern_iftype in InterfacesRegex:
            pattern = pattern_iftype.value
            if re.match(pattern, part_num):
                return getattr(NetboxInterfaceTypes, pattern_iftype.name).value

        raise TypeCouldNotBeParsedError()

    def _get_transceiver_by_if(self):
        cmd = "show interface transceiver" + self._driver_end_selection()

        if not self.cache.get("transceivers"):
            transceiver_conf_dump = self.device.cli([cmd])[cmd]
            transceivers = self._correct_and_convert_to_dict(
                transceiver_conf_dump)["TABLE_interface"]["ROW_interface"]

            self.cache["transceivers"] = {
                i["interface"]: i for i in transceivers
            }

        return self.cache["transceivers"]

    def _guess_type_from_if_type(self, interface):
        from pynxos.errors import CLIError

        try:
            status = self._get_ifstatus_by_if()[interface]
            if_speed, if_type = status["speed"], status["type"]
        except (json.JSONDecodeError, KeyError, CLIError):
            raise TypeCouldNotBeParsedError()

        if not if_type.strip("-"):
            try:
                if int(if_speed) == 1000:
                    return NetboxInterfaceTypes.eth1000.value
            except ValueError:
                pass
        else:
            for pattern in InterfacesRegex:
                if re.match(pattern.value, if_type):
                    return getattr(NetboxInterfaceTypes, pattern.name).value

        raise TypeCouldNotBeParsedError()

    def _get_ifstatus_by_if(self):
        cmd = "show interface status" + self._driver_end_selection()

        if not self.cache.get("ifstatus"):
            status_conf_dump = self.device.cli([cmd])[cmd]
            status = self._correct_and_convert_to_dict(status_conf_dump)[
                "TABLE_interface"]["ROW_interface"]

            self.cache["ifstatus"] = {
                i["interface"]: i for i in status
            }

        return self.cache["ifstatus"]

    def get_detailed_lldp_neighbours(self):
        """
        Napalm does not show id for neighbours. Gives a little more info

        :return neighbours: [{
                "local_port": local port name,
                "hostname": neighbour hostname (if handled),
                "port": neighbour port name,
                "chassis_id": neighbour id
            }]
        """
        cmd = "show lldp neighbors detail" + self._driver_end_selection()

        cmd_output = self.device.cli([cmd])[cmd]
        neighbours = self._correct_and_convert_to_dict(cmd_output)[
            "TABLE_nbor_detail"]["ROW_nbor_detail"]

        if isinstance(neighbours, dict):
            neighbours = [neighbours]

        for n in neighbours:
            yield {
                "local_port": n["l_port_id"],
                "hostname": n["sys_name"],
                "port": n["port_id"],
                "chassis_id": n["chassis_id"]
            }

    def get_detailed_cdp_neighbours(self):
        """
        Napalm does not support cdp

        :return neighbours: [{
                "local_port": local port name,
                "hostname": neighbour hostname (if handled),
                "port": neighbour port name,
            }]
        """
        cmd = "show cdp neighbors detail" + self._driver_end_selection()

        cmd_output = self.device.cli([cmd])[cmd]
        neighbours = self._correct_and_convert_to_dict(cmd_output)[
            "TABLE_cdp_neighbor_detail_info"]["ROW_cdp_neighbor_detail_info"]

        if isinstance(neighbours, dict):
            neighbours = [neighbours]

        for n in neighbours:
            yield {
                "local_port": n["intf_id"],
                "hostname": re.sub("\(.*$", "", n["device_id"], count=1),
                "port": n["port_id"],
            }

    def _driver_end_selection(self) -> str:
        if type(self.device) is NXOSDriver:
            self.device.device.api.cmd_method_raw = "cli"
            return ""
        else:
            return " | json"

    def _correct_and_convert_to_dict(self, cmd_output) -> dict:
        if type(cmd_output) is not dict:
            if not re.search("^{", cmd_output):
                cmd_output = re.sub('^.+\. {', '{', cmd_output, count=1)
            cmd_output = json.loads(cmd_output)
        return cmd_output

    def get_interfaces_lag(self, interfaces):
        cmd = "show port-channel summary" + self._driver_end_selection()
        interfaces_lag = defaultdict(list)

        cmd_output = self.device.cli([cmd])[cmd]
        port_cannels = self._correct_and_convert_to_dict(cmd_output)[
            "TABLE_channel"]["ROW_channel"]

        if isinstance(port_cannels, dict):
            port_cannels = [port_cannels]

        for p in port_cannels:
            if not p.get("TABLE_member"):
                continue
            if not p["TABLE_member"].get("ROW_member"):
                continue
            if isinstance(p["TABLE_member"]["ROW_member"], dict):
                member = [p["TABLE_member"]["ROW_member"]]
            else:
                member = p["TABLE_member"]["ROW_member"]
            for interface in member:
                interfaces_lag[interface["port"]] = p["port-channel"]

        return interfaces_lag

    def get_interface_mode(self, interface):
        from pynxos.errors import CLIError
        try:
            return self._get_interfaces_mode()[interface].get("oper_mode")
        except (KeyError, CLIError):
            logger.debug("Switch %s, show interface switchport cmd error",
                         self.device.hostname)
        return None

    def get_interface_access_vlan(self, interface):
        from pynxos.errors import CLIError
        try:
            return int(
                self._get_interfaces_mode()[interface].get("access_vlan")
            )
        except (KeyError, CLIError):
            logger.debug("Switch %s, show interface switchport cmd error",
                         self.device.hostname)
        return None

    def get_interface_netive_vlan(self, interface):
        from pynxos.errors import CLIError
        try:
            return int(
                self._get_interfaces_mode()[interface].get("native_vlan")
            )
        except (KeyError, CLIError):
            logger.debug("Switch %s, show interface switchport cmd error",
                         self.device.hostname)
        return None

    def _get_interfaces_mode(self):
        cmd = "show interface switchport" + self._driver_end_selection()

        if not self.cache.get("mode"):
            mode_conf_dump = self.device.cli([cmd])[cmd]
            mode = self._correct_and_convert_to_dict(mode_conf_dump)[
                "TABLE_interface"]["ROW_interface"]

            self.cache["mode"] = {
                i["interface"]: i for i in mode
            }

        return self.cache["mode"]

    def get_vlans(self):
        """
        Napalm does not support vlan
        but there are issues https://github.com/napalm-automation/napalm/issues/927

        :return vlans:
            vlan_id, {
                "name": vlan name,
                "interfaces": list interfaces dict
            }
        """
        cmd = "show vlan brief" + self._driver_end_selection()

        cmd_output = self.device.cli([cmd])[cmd]
        vlans = self._correct_and_convert_to_dict(cmd_output)[
            "TABLE_vlanbriefxbrief"]["ROW_vlanbriefxbrief"]

        if isinstance(vlans, dict):
            vlans = [vlans]

        for v in vlans:
            yield v["vlanshowbr-vlanid"], {
                "name": v["vlanshowbr-vlanname"],
                "interfaces": self._parse_ports(v["vlanshowplist-ifidx"])
            }

    def _parse_ports(self, vlan_s) -> list:
        vlans = []
        find_regexp = r"^([A-Za-z\/-]+|.*\/)(\d+)-(\d+)$"
        vlan_str = ""

        if isinstance(vlan_s, list):
            for v in vlan_s:
                vlan_str += "," + v
        else:
            vlan_str = vlan_s

        for vls in vlan_str.split(","):
            find = re.findall(find_regexp, vls.strip())
            if find:
                for i in range(int(find[0][1]), int(find[0][2]) + 1):
                    vlans.append(find[0][0] + str(i))
            else:
                vlans.append(vls.strip())
        return vlans



from collections import defaultdict
import logging
import defusedxml.lxml
import lxml.etree
# should be a napalm dependency
from jnpr.junos.exception import RpcError
import re

from netbox_netprod_importer.vendors import _AbstractVendorParser
from netbox_netprod_importer.vendors.constants import NetboxInterfaceTypes
from netbox_netprod_importer.exceptions import TypeCouldNotBeParsedError
from .constants import InterfacesRegex


logger = logging.getLogger("netbox_importer")


class JuniperParser(_AbstractVendorParser):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cache = {}

    def get_interfaces_lag(self, interfaces):
        return super().get_interfaces_lag(interfaces)

    def get_interface_type(self, interface):
        return super().get_interface_type(interface)

    @staticmethod
    def get_real_ifname(interface):
        ifsplit = interface.split(".")
        if len(ifsplit) > 1 and ifsplit[0].lower() != "vlan":
            return ifsplit[0]
        else:
            return interface


class JunOSParser(JuniperParser):

    def get_interfaces_lag(self, interfaces):
        interfaces_lag = defaultdict(set)

        try:
            interfaces_info_xml = self.device._rpc(
                self._gen_rpc_request_interfaces_info()
            )
        except RpcError as e:
            logger.debug("RPC error: %s", e)
            raise

        parsed_xml = defusedxml.lxml.fromstring(interfaces_info_xml)

        for ifblock in parsed_xml.xpath(".//physical-interface"):
            ifname = ifblock.xpath("name")[0].text.strip()
            if ifname not in interfaces:
                continue

            bundle_xpath = ifblock.xpath(".//ae-bundle-name")
            if bundle_xpath:
                bundle_name = bundle_xpath[0].text.split(".")[0].strip()
                interfaces_lag[ifname] = bundle_name

        return interfaces_lag

    def _gen_rpc_request_interfaces_info(self):
        get_pic_details_el = lxml.etree.Element("get-interface-information")
        xml_tree = get_pic_details_el.getroottree()

        return lxml.etree.tostring(xml_tree).decode()

    def get_interface_type(self, interface):
        try:
            junos_type = self._guess_type_from_chassis_pic(interface)
        except TypeCouldNotBeParsedError:
            return "Other"

        for pattern_iftype in InterfacesRegex:
            pattern = pattern_iftype.value
            if re.match(pattern, junos_type):
                return getattr(NetboxInterfaceTypes, pattern_iftype.name).value

        return "Other"

    def _guess_type_from_chassis_pic(self, interface):
        pattern = r".*-(\d+)/(\d+)/(\d+)"
        matching_ifname_split = re.match(pattern, interface)

        if not matching_ifname_split:
            logger.debug("Unexpected interface name convention: %s", interface)
            raise TypeCouldNotBeParsedError(interface)

        fpc, pic, port_index = map(int, matching_ifname_split.groups())
        if max((fpc, pic, port_index)) > 254:
            return "Other"

        if "pic" not in self.cache:
            self.cache["pic"] = {}

        if (fpc, pic) not in self.cache["pic"]:
            try:
                chassis_pic_xml = self.device._rpc(
                    self._gen_rpc_request_pic_info(fpc, pic)
                )
            except RpcError as e:
                logger.debug("RPC error: %s", e)
                raise TypeCouldNotBeParsedError(interface)

            parsed_xml = defusedxml.lxml.fromstring(chassis_pic_xml)
            self.cache["pic"][(fpc, pic)] = parsed_xml
        else:
            parsed_xml = self.cache["pic"][(fpc, pic)]

        try:
            ports_info = parsed_xml.findall(".//port-information")[0]
        except IndexError:
            return parsed_xml.findall(".//pic-type")[0].text

        for p in ports_info.getchildren():
            if int(p.xpath("port-number")[0].text) == port_index:
                return p.xpath("sfp-vendor-pno")[0].text

        raise TypeCouldNotBeParsedError(interface)

    def _gen_rpc_request_pic_info(self, fpc, pic):
        get_pic_details_el = lxml.etree.Element("get-pic-detail")
        xml_tree = get_pic_details_el.getroottree()

        for slot_type in ("fpc", "pic"):
            slot_el = lxml.etree.Element("{}-slot".format(slot_type))
            get_pic_details_el.append(slot_el)
            slot_el.text = str(vars()[slot_type])

        return lxml.etree.tostring(xml_tree).decode()

    def get_detailed_lldp_neighbours(self):
        try:
            lldp_neighbours_xml = self.device._rpc(
                self._gen_rpc_lldp_neighbours()
            )
        except RpcError as e:
            logger.debug("RPC error: %s", e)
            raise

        parsed_xml = defusedxml.lxml.fromstring(lldp_neighbours_xml)

        for n in parsed_xml.xpath(".//lldp-neighbor-information"):
            yield {
                "local_port": (
                    n.xpath(
                        "lldp-local-interface"
                    )[0].text.split(".")[0].strip()
                ),
                "hostname": (
                    n.xpath("lldp-remote-system-name")[0].text.strip()
                ),
                "port": (
                    n.xpath(
                        "lldp-remote-port-description"
                    )[0].text.split(".")[0].strip()
                ),
                "chassis_id": (
                    n.xpath("lldp-remote-chassis-id")[0].text.strip()
                ),
            }

    def _gen_rpc_lldp_neighbours(self):
        get_lldp_neighbours_el = lxml.etree.Element(
            "get-lldp-neighbors-information"
        )
        xml_tree = get_lldp_neighbours_el.getroottree()

        return lxml.etree.tostring(xml_tree).decode()

    # The following features are stubs. Since I can not test.
    def get_interface_mode(self, interface):
        logger.info("get_interface_mode function not implemented for junos")
        raise NotImplementedError

    def get_interface_access_vlan(self, interface):
        logger.info("get_interface_access_vlan function not implemented for "
                    "junos")
        raise NotImplementedError

    def get_interface_netive_vlan(self, interface):
        logger.info("get_interface_netive_vlan function not implemented for "
                    "junos")
        raise NotImplementedError

    def get_interface_vlans(self, interface):
        logger.info("get_interface_vlans function not implemented for junos")
        raise NotImplementedError



import cachetools
import re

from netbox_netprod_importer.vendors import _AbstractVendorParser


class CiscoParser(_AbstractVendorParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.cache = cachetools.TTLCache(10, 60)

    @staticmethod
    def get_abrev_if(interface):
        if_index_re = re.search(r"\d.*", interface)
        if_index_re = if_index_re.group() if if_index_re else ""

        if interface.lower().startswith("eth"):
            prefix = interface[:3]
        else:
            prefix = interface[:2]

        return prefix + if_index_re

    def get_interface_vlans(self, interface):

        if not self.cache.get("vlan"):
            self.cache["ttl"] = 600
            self.cache["vlan"] = {}
            for vlan, data in self.get_vlans():
                for iface in data["interfaces"]:
                    if not self.cache["vlan"].get(iface):
                        self.cache["vlan"][iface] = []
                    self.cache["vlan"][iface].append(vlan)
        return self.cache["vlan"].get(interface)





import appdirs
import errno
import logging
import os
import yaml
import urllib3

from netbox_netprod_importer import __appname__


logger = logging.getLogger("virt_backup")

os.environ["XDG_CONFIG_DIRS"] = "/etc"
CONFIG_DIRS = (
    appdirs.user_config_dir(__appname__),
    appdirs.site_config_dir(__appname__),
)
CONFIG_FILENAME = "config.yml"


def get_config(custom_path=None):
    """
    Get config file and load it with yaml
    :returns: loaded config in yaml, as a dict object
    """
    if getattr(get_config, "cache", None):
        return get_config.cache

    if custom_path:
        config_path = custom_path
    elif os.environ.get("CONFIG_PATH"):
        config_path = os.environ["CONFIG_PATH"]
    else:
        for d in CONFIG_DIRS:
            config_path = os.path.join(d, CONFIG_FILENAME)
            if os.path.isfile(config_path):
                break
    try:
        with open(config_path, "r") as config_file:
            conf = yaml.safe_load(config_file)
            get_config.cache = conf
            if conf.get("loglevel", "None"):
                numeric_level = getattr(logging, conf.get("loglevel").upper())
                if not isinstance(numeric_level, int):
                    raise ValueError('Invalid log level: %s' \
                                     % conf.get("loglevel"))
            else:
                numeric_level = logging.ERROR
            logging.basicConfig(level=numeric_level,
                                format="%(levelname)s: %(name)s: %(message)s"
                                )
            if conf.get("disable_ssl_warnings", False):
                urllib3.disable_warnings()
            return conf
    except FileNotFoundError as e:
        logger.debug(e)
        if custom_path or os.environ.get("CONFIG_PATH"):
            logger.error(
                "Configuration file {} not found.".format(
                    custom_path or os.environ["CONFIG_PATH"]
                )
            )
        else:
            logger.error(
                "No configuration file can be found. Please create a "
                "config.yml in one of these directories:\n"
                "{}".format(", ".join(CONFIG_DIRS))
            )
        raise FileNotFoundError


def load_config(custom_path=None):
    get_config(custom_path)



from collections import defaultdict
from contextlib import ContextDecorator
import logging
import socket
import napalm

from netbox_netprod_importer.exceptions import (
    NoReverseFoundError, DeviceNotSupportedError
)
from netbox_netprod_importer.vendors import DeviceParsers, StubParser
from netbox_netprod_importer.tools import is_macaddr

logger = logging.getLogger("netbox_importer")


class DeviceImporter(ContextDecorator):

    def __init__(self, hostname, napalm_driver_name, target=None, creds=None,
                 napalm_optional_args=None, discovery_protocol='lldp'):
        self.hostname = hostname
        if not creds:
            creds = (None, None)
        self.target = target or hostname

        driver = napalm.get_network_driver(napalm_driver_name)
        self.device = driver(
            hostname=self.target, username=creds[0], password=creds[1],
            optional_args=napalm_optional_args
        )
        self.specific_parser = self._get_specific_device_parser(
            napalm_driver_name
        )
        self.discovery_protocol = discovery_protocol
        self.napalm_driver_name = napalm_driver_name

    def _get_specific_device_parser(self, os):
        try:
            parser_class = getattr(DeviceParsers, os).value
        except AttributeError:
            logger.info(
                "{} is not totally supported, will be limited to napalm "
                "features"
            )
            parser_class = StubParser

        return parser_class(self.device)

    def __enter__(self):
        return self.open()

    def __exit__(self, *exc):
        if self.device.device:
            self.close()

    def open(self):
        self.device.open()

    def close(self):
        self.device.close()

    def poll(self):
        assert self.device.device

        props = {}

        logging.debug("Trying to resolve the primaries IP")
        try:
            props.update(self.resolve_primary_ip())
        except NoReverseFoundError:
            logger.error(
                "Cannot fill primary ip for host %s, no reverse found.",
                self.hostname
            )

        try:
            props.update(self._handle_serial_num())
        except DeviceNotSupportedError:
            logger.error(
                "Cannot fetch serial, device %s not supported", self.hostname
            )

        self._handle_interfaces_props(props)

        return props

    def resolve_primary_ip(self):
        """
        Resolve primary IPs from hostname

        :return: {"primary_ipv4": ipv4, "primary_ipv6": ipv6}, each key will
                 exist if a reverse exists for this host
        """
        main_ip = {}

        assoc_proto_socket = (
            ("primary_ip4", socket.AF_INET), ("primary_ip6", socket.AF_INET6)
        )
        for proto, socket_type in assoc_proto_socket:
            try:
                main_ip[proto] = socket.getaddrinfo(
                    self.hostname, None, socket_type
                )[0][4][0]
            except socket.gaierror as e:
                logger.debug(
                    "Error resolving a reverse for %s for family %s: %s",
                    self.hostname, socket_type, e
                )

        if not main_ip:
            raise NoReverseFoundError(self.hostname)

        return main_ip

    def _handle_serial_num(self):
        assert self.device.device

        try:
            serial = self.device.get_facts()["serial_number"]
        except IndexError:
            raise DeviceNotSupportedError(self.hostname)

        return {"serial": serial} if serial else {}

    def _handle_interfaces_props(self, props):
        logger.debug("Get properties of each interface")
        interfaces = self.get_interfaces()
        logger.debug("Get IP setup on each interface")
        interfaces = self.fill_interfaces_ip(interfaces)

        props["interfaces"] = interfaces
        return props

    def get_interfaces(self):
        assert self.device.device

        napalm_interfaces = self.device.get_interfaces()

        interfaces = {}
        trunks = []
        # sort to test it more easily
        for ifname, napalm_ifprops in sorted(napalm_interfaces.items()):
            is_subif, parent_if = self._is_subinterface(ifname)
            if is_subif:
                trunks.append(parent_if)
                continue

            if not is_macaddr(napalm_ifprops["mac_address"]):
                napalm_ifprops["mac_address"] = None

            _type = self.specific_parser.get_interface_type(ifname)
            if _type == "Other":
                logger.info("Switch %s iface type for the port %s not defined.",
                            self.hostname, ifname)
            try:
                mode = {
                    "access": "Access",
                    "trunk": "Tagged",
                    "static access": "Access",
                    None: None
                }[self.specific_parser.get_interface_mode(ifname)]
            except NotImplementedError:
                mode = None
            except KeyError as ex:
                logger.debug("Switch %s Interface %s unknown mode: %s.",
                             self.hostname, ifname, ex)
                mode = None

            interfaces[ifname] = {
                "enabled": napalm_ifprops["is_enabled"],
                # Netbox max descr size is 100 char
                "description": (
                    napalm_ifprops["description"] or ""
                )[:100],
                "mac_address": napalm_ifprops["mac_address"] or None,
                # wait for this pull request
                # https://github.com/napalm-automation/napalm/pull/531
                "mtu": napalm_ifprops.get("mtu", None),
                "type": _type,
                "mode": mode,
                "untagged_vlan": None,
                "tagged_vlans": [],
            }

            if mode == "Access":
                try:
                    interfaces[ifname]["untagged_vlan"] = \
                        self.specific_parser.get_interface_access_vlan(ifname)
                except NotImplementedError:
                    interfaces[ifname]["untagged_vlan"] = None

            try:
                vlans = self.specific_parser.get_interface_vlans(ifname)
            except NotImplementedError:
                vlans = None

            if vlans:
                interfaces[ifname]["tagged_vlans"] = vlans


        for ifname, data in interfaces.items():
            if data["mode"] == "Tagged":
                try:
                    navive = self.specific_parser.get_interface_netive_vlan(ifname)
                except NotImplementedError:
                    navive = None
                if navive in data["tagged_vlans"]:
                    interfaces[ifname]["untagged_vlan"] = navive
                    interfaces[ifname]["tagged_vlans"].pop(navive)

        for trunk in trunks:
            if trunk in interfaces:
                interfaces[trunk]["mode"] = "Tagged"

        interfaces_lag = self.specific_parser.get_interfaces_lag(interfaces)
        for ifname, lag in interfaces_lag.items():
            try:
                real_lag_name = (
                    self._search_key_case_insensitive(interfaces, lag)
                )
            except KeyError:
                logger.error("%s not exist in polled interfaces", lag)
                continue

            interfaces[ifname]["lag"] = real_lag_name
            interfaces[real_lag_name]["type"] = (
                "Link Aggregation Group (LAG)"
            )

        return interfaces

    def _is_subinterface(self, interface):
        ifsplit = interface.split(".")
        if len(ifsplit) > 1 and ifsplit[0].lower() != "vlan":
            return True, ifsplit[0]
        else:
            return False, interface

    def _search_key_case_insensitive(self, dictionary, key):
        if key in dictionary:
            return key

        for k in dictionary.keys():
            if k == key or isinstance(key, str) and k.lower() == key.lower():
                return k

        raise KeyError()

    def fill_interfaces_ip(self, interfaces=None):
        assert self.device.device

        if interfaces is None:
            interfaces = defaultdict(dict)
            skip_unlisted_if = False
        else:
            skip_unlisted_if = True

        for ifname, ifprops in self.device.get_interfaces_ip().items():
            is_subif, parent_if = self._is_subinterface(ifname)
            if is_subif:
                ifname = parent_if

            if skip_unlisted_if and ifname not in interfaces:
                logger.debug(
                    "Interface {} has IP but was not listed".format(ifname)
                )
                continue

            if not interfaces[ifname].get("ip"):
                interfaces[ifname]["ip"] = []

            interfaces[ifname]["ip"].extend(
                "{}/{}".format(ip, ip_props["prefix_length"])
                for proto in ("ipv4", "ipv6")
                if ifprops.get(proto, None)
                for ip, ip_props in ifprops[proto].items()
            )

        return interfaces

    def get_neighbours(self):
        """
        Either try the specific way to get lldp, cdp neighbours, or try using napalm
        if not supported

        :return neighbours: [{
                "local_port": local port name,
                "hostname": neighbour hostname (if handled),
                "port": neighbour port name,
                "mgmt_id": neighbour id # only with specific parser
            }]
        """
        assert self.device.device
        if self.discovery_protocol == "cdp":
            yield from self.get_cdp_neighbours()
        elif self.discovery_protocol == "multiple" and \
                self.napalm_driver_name in ['ios', 'nxos', 'nxos_ssh']:
            yield from self.get_multiple_neighbours()
        else:
            yield from self.get_lldp_neighbours()

    def get_multiple_neighbours(self):
        neighbours = [n for n in self.get_cdp_neighbours()]
        for n in self.get_lldp_neighbours():
            if not any(self.specific_parser.get_abrev_if(elem['local_port']) \
                       == self.specific_parser.get_abrev_if(n['local_port']) \
                       for elem in neighbours):
                yield n
        for n in neighbours:
            yield n

    def get_cdp_neighbours(self):
        """
        Either try the specific way to get cdp neighbours

        :return neighbours: [{
                "local_port": local port name,
                "hostname": neighbour hostname (if handled),
                "port": neighbour port name,
            }]
        """
        try:
            yield from self.specific_parser.get_detailed_cdp_neighbours()
        except NotImplementedError:
            logger.error("%s platform does not support cdp", self.platform)
            return []

    def get_lldp_neighbours(self):
        """
        Either try the specific way to get lldp neighbours, or try using napalm
        if not supported

        :return neighbours: [{
                "local_port": local port name,
                "hostname": neighbour hostname (if handled),
                "port": neighbour port name,
                "mgmt_id": neighbour id # only with specific parser
            }]
        """

        try:
            yield from self.specific_parser.get_detailed_lldp_neighbours()
        except NotImplementedError:
            napalm_neighbours = self.device.get_lldp_neighbors()
            for port, port_neighbours in napalm_neighbours.items():
                for neighbour in port_neighbours:
                    yield {
                        "local_port": port,
                        "hostname": neighbour["hostname"],
                        "port": neighbour["port"]
                    }



from abc import ABC, abstractmethod
import logging


logger = logging.getLogger("netbox_importer")


class _AbstractVendorParser(ABC):

    def __init__(self, napalm_device, *args, **kwargs):
        self.device = napalm_device

    @abstractmethod
    def get_interfaces_lag(self, interfaces):
        logger.debug("Get interfaces LAG on host %s", self.device.hostname)
        return {}

    @abstractmethod
    def get_interface_type(self, interface):
        logger.debug(
            "Get type of interface %s on host %s",
            interface, self.device.hostname
        )

    def get_all_derivatives_for_netif(self, interface):
        """
        Get all possible derivatives for an interface name
        """
        yield interface

    def get_detailed_lldp_neighbours(self):
        """
        Napalm does not show id for neighbours. Gives a little more info

        :return neighbours: [{
                "local_port": local port name,
                "hostname": neighbour hostname (if handled),
                "port": neighbour port name,
                "mgmt_id": neighbour id
            }]
        """
        raise NotImplementedError()



from abc import ABC, abstractmethod
from collections import defaultdict
import concurrent.futures
from concurrent.futures import ThreadPoolExecutor
import logging
from requests.exceptions import HTTPError
import threading

import cachetools
from netboxapi import NetboxMapper
from tqdm import tqdm

from netbox_netprod_importer.vendors.cisco import CiscoParser
from netbox_netprod_importer.vendors.juniper import JuniperParser
from netbox_netprod_importer.exceptions import (
    DeviceNotFoundError, IPPushingError, NetInterfaceNotFoundError,
    NetIfPushingError
)
from netbox_netprod_importer.tools import (
    generic_netbox_error, is_macaddr, macaddr_to_int
)


logger = logging.getLogger("netbox_importer")


class _NetboxPusher(ABC):

    def __init__(self, netbox_api, *args, **kwargs):
        self.netbox_api = netbox_api

        self._mappers = {
            "dcim_choices": NetboxMapper(
                self.netbox_api, app_name="dcim", model="_choices"
            ), "devices": NetboxMapper(
                self.netbox_api, app_name="dcim", model="devices"
            ), "interfaces": NetboxMapper(
                self.netbox_api, app_name="dcim", model="interfaces"
            ), "cables": NetboxMapper(
                self.netbox_api, app_name="dcim", model="cables"
            ), "ip": NetboxMapper(
                self.netbox_api, app_name="ipam", model="ip-addresses"
            ), "vlan": NetboxMapper(
                self.netbox_api, app_name="ipam", model="vlans"
            )
        }
        self._choices_cache = {}

    @abstractmethod
    def push(self):
        pass

    def search_value_in_choices(self, mapper_name, id, label):
        if mapper_name not in self._choices_cache:
            try:
                mapper = self._mappers[mapper_name]
                self._choices_cache[mapper_name] = next(mapper.get())
            except StopIteration:
                pass

        for choice in self._choices_cache[mapper_name][id]:
            if choice["label"] == label:
                return choice["value"]

        raise KeyError("Label {} not in choices".format(label))


class NetboxDevicePropsPusher(_NetboxPusher):
    _device = None

    def __init__(self, netbox_api, hostname, props, *args, overwrite=False,
                 **kwargs):
        super().__init__(netbox_api, *args, **kwargs)

        self.hostname = hostname
        self.props = props
        self.overwrite = overwrite
        self.vlans_cache = cachetools.LRUCache(4096)

    @generic_netbox_error
    def push(self):
        try:
            self._device = next(
                self._mappers["devices"].get(name=self.hostname)
            )
        except StopIteration:
            raise DeviceNotFoundError(self.hostname)

        if self.overwrite:
            self._clean_unmatched_interfaces()
        self._push_interfaces()
        self._push_main_data()

    def _clean_unmatched_interfaces(self):
        # Interfaces are all forced to be fetched, as some of them will them
        # be deleting, messing with the offset used by the query to fetch the
        # next pool.
        pushed_interfaces = tuple(
            self._mappers["interfaces"].get(device_id=self._device)
        )

        for netbox_if in pushed_interfaces:
            if netbox_if.name not in self.props["interfaces"]:
                self._clean_attached_ip(netbox_if)
                netbox_if.delete()

    def _clean_attached_ip(self, netbox_if):
        attached_addrs = self._mappers["ip"].get(interface_id=netbox_if)
        for a in attached_addrs:
            a.delete()

    def _push_interfaces(self):
        interfaces_props = self.props["interfaces"]
        interfaces_lag = {}
        interfaces = {}

        for if_name, if_prop in interfaces_props.items():
            if_prop = if_prop.copy()
            interface_query = self._mappers["interfaces"].get(
                device_id=self._device, name=if_name
            )
            try:
                interface = next(interface_query)
            except StopIteration:
                try:
                    interface = self._mappers["interfaces"].post(
                        device=self._device, name=if_name
                    )
                except HTTPError as e:
                    raise NetIfPushingError(if_name, e)

            interfaces[if_name] = interface

            if_type = if_prop.pop("type")
            if_prop["form_factor"] = self.search_value_in_choices(
                "dcim_choices", "interface:type",
                if_type
            )
            if if_prop.get("lag"):
                interfaces_lag[if_name] = if_prop.pop("lag")

            if not self.overwrite and "mode" in if_prop:
                if_prop.pop("mode")
            elif if_prop.get("mode"):
                # cannot really guess (yet) the interface mode, so only set it
                # if overwrite
                self._handle_interface_mode(interface, if_prop["mode"])
                if_prop.pop("mode")

            if if_prop["untagged_vlan"]:
                vlan_id = self._get_vlan_id(if_prop["untagged_vlan"])
                if vlan_id != -1:
                    setattr(interface, "untagged_vlan", vlan_id)

            if_prop.pop("untagged_vlan")
            if len(if_prop["tagged_vlans"]):
                setattr(interface, "tagged_vlans", [])
                for vlan in if_prop["tagged_vlans"]:
                    vlan_id = self._get_vlan_id(vlan)
                    if vlan_id != -1:
                        interface.tagged_vlans.append(vlan_id)
            if_prop.pop("tagged_vlans")

            for k, v in if_prop.items():
                setattr(interface, k, v)

            try:
                interface.put()
            except HTTPError as e:
                raise NetIfPushingError(interface.name, e)
            if if_prop.get("ip"):
                addrs = self._attach_interface_to_ip_addresses(
                    interface, *if_prop["ip"]
                )
                if self.overwrite:
                    self._clean_unmatched_ip_addresses(interface, *addrs)

        self._update_interfaces_lag(interfaces, interfaces_lag)

    def _get_vlan_id(self, vlan):
        if not self.vlans_cache.get(self._device.site.id):
            self.vlans_cache[self._device.site.id] = {}

        if not self.vlans_cache[self._device.site.id].get(vlan):
            vlans = list(self._mappers["vlan"].get(
                    site_id=self._device.site.id,
                    vid=vlan
                )
            )
            # Ignore vlan 1 because it is usually not used.
            # But if he is assign.
            if len(vlans) == 0 and vlan != 1:
                logger.info("Switch %s, vlan %s not faund on site %s",
                            self.hostname, vlan, self._device.site.name)
                # If set to None, then if not None will always trigger.
                # Searches will occur every time.
                self.vlans_cache[self._device.site.id][vlan] = -1
            elif len(vlans) == 1:
                self.vlans_cache[self._device.site.id][vlan] = vlans[0].id
            else:
                logger.info(
                    "Number of found Vlans %s on the site %s is more than one",
                    vlan, self._device.site.name
                )
                self.vlans_cache[self._device.site.id][vlan] = -1
        return self.vlans_cache[self._device.site.id][vlan]


    def _handle_interface_mode(self, netbox_if, mode):
        netbox_mode = self.search_value_in_choices(
            "dcim_choices", "interface:mode",
            mode
        )

        netbox_if.mode = netbox_mode

    def _attach_interface_to_ip_addresses(self, netbox_if, *ip_addresses):
        mapper = self._mappers["ip"]

        addresses = []
        for ip in ip_addresses:
            # check if ip attached isn't already correct
            try:
                ip_netbox_obj = next(mapper.get(q=ip, interface_id=netbox_if))
            except StopIteration:
                try:
                    ip_netbox_obj = next(mapper.get(q=ip))
                except StopIteration:
                    try:
                        ip_netbox_obj = mapper.post(address=ip)
                    except HTTPError as e:
                        raise IPPushingError(ip, e)

                # XXX: handle anycast
                ip_netbox_obj.interface = netbox_if
                try:
                    ip_netbox_obj.put()
                except HTTPError as e:
                    raise IPPushingError(ip_netbox_obj.address, e)

            addresses.append(ip_netbox_obj)

        return addresses

    def _clean_unmatched_ip_addresses(self, netbox_if, *netbox_ip):
        attached_addrs = self._mappers["ip"].get(interface_id=netbox_if)
        netbox_ip_ids = set(obj.id for obj in netbox_ip)

        for addr in attached_addrs:
            if addr.id not in netbox_ip_ids:
                addr.delete()

    def _update_interfaces_lag(self, interfaces, interfaces_lag):
        """
        :param interfaces: {interface_name: netbox_interface_obj, …}
        """
        for if_name, lag in interfaces_lag.items():
            interface = interfaces[if_name]
            interface.lag = interfaces[lag]
            try:
                interface.put()
            except HTTPError as e:
                raise NetIfPushingError(interface.name, e)

    def _push_main_data(self):
        mapper = self._mappers["ip"]
        if self.props.get("serial"):
            self._device.serial = self.props["serial"]

        for ip_key in ("primary_ip4", "primary_ip6"):
            ip = self.props.get(ip_key)
            if ip:
                try:
                    ip_netbox_obj = next(mapper.get(q=ip))
                    setattr(self._device, ip_key, ip_netbox_obj)
                except StopIteration:
                    logger.error(
                        "Cannot set primary IP %s as it does not exist in "
                        "netbox", ip
                    )

        self._device.put()


class NetboxInterconnectionsPusher(_NetboxPusher):
    """
    Push in Netbox a graph representing the interconnections between devices
    """

    def __init__(self, *args, remove_domains=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.remove_domains = remove_domains or []
        self.interfaces_cache = cachetools.LRUCache(128)
        self._lock = threading.Lock()

    def push(self, importers, threads=1, overwrite=False):
        result = {"done": 0, "errors_interco": 0, "errors_device": 0}

        importers = importers.copy()
        with ThreadPoolExecutor(max_workers=threads) as executor:
            futures = {}
            discovered = defaultdict(dict)
            for host, importer in importers.items():
                future = executor.submit(
                    self._handle_device, host, importer, discovered, overwrite
                )
                futures[future] = host

            futures_with_progress = tqdm(
                concurrent.futures.as_completed(futures),
                total=len(futures)
            )
            for future in futures_with_progress:
                host = futures[future]
                try:
                    task_result = future.result()
                    result["done"] += task_result["done"]
                    result["errors_interco"] += task_result["errors"]
                except ValueError:
                    logger.debug(
                        "LLDP parsing not supported on {}".format(host)
                    )
                    result["errors_device"] += 1
                except Exception as e:
                    logger.debug(
                        "Error when defining interconnections on host %s: %s",
                        host, e
                    )
                    result["errors_device"] += 1
                importers.pop(host)

        return result

    def _handle_device(self, hostname, importer, discovered, overwrite):
        result = {"done": 0, "errors": 0}
        with importer:
            for interco in importer.get_neighbours():
                already_discovered = (
                    discovered[importer.hostname].get(
                        interco["local_port"], None
                    ) == (interco["hostname"], interco["port"])
                )
                if already_discovered:
                    continue

                try:
                    discovered[importer.hostname][interco["local_port"]] = (
                        interco["hostname"], interco["port"]
                    )
                    discovered[interco["hostname"]][interco["port"]] = (
                        importer.hostname, interco["local_port"]
                    )

                    try:
                        netif_connection = self._interconnect_using_lldp_names(
                            hostname, importer, interco
                        )
                    except DeviceNotFoundError:
                        if "chassis_id" not in interco:
                            raise

                        netif_connection = self._interconnect_using_lldp_id(
                            hostname, importer, interco
                        )

                    self._update_discovered_from_netif_connection(
                        discovered, netif_connection
                    )

                    result["done"] += 1
                    logger.debug("True with interco %s:", interco)
                except Exception as e:
                    result["errors"] += 1
                    logger.warning("Switch %s Error with interco %s: %s",
                                   hostname, interco, e)
                    continue

        if overwrite:
            self._clean_undetected_intercos(hostname, discovered)

        return result

    @generic_netbox_error
    def _interconnect_using_lldp_names(self, hostname, importer, interco):
        a = hostname
        netif_a = self._get_netif_or_derivative(a, interco["local_port"])
        b = interco["hostname"]

        for dom in self.remove_domains:
            dom = "." + dom.lstrip(".").rstrip(".")
            if b.endswith(dom):
                b = b.rstrip(dom)
                break

        netif_b = self._get_netif_or_derivative(b, interco["port"])

        with self._lock:
            # force a refresh of the interfaces
            netif_a = next(netif_a.get())
            netif_b = next(netif_b.get())
            return self.interconnect_netbox_netif(netif_a, netif_b)

    @generic_netbox_error
    def _interconnect_using_lldp_id(self, hostname, importer, interco):
        a = hostname
        netif_a = self._get_netif_or_derivative(a, interco["local_port"])
        netif_b = self._find_netbox_netif_from_lldp_id(
            interco["chassis_id"], interco["port"]
        )

        with self._lock:
            # force a refresh of the interfaces
            netif_a = next(netif_a.get())
            netif_b = next(netif_b.get())
            return self.interconnect_netbox_netif(netif_a, netif_b)

    def _find_netbox_netif_from_lldp_id(self, lldp_id, if_name):
        """
        Find an interface in netbox from a LLDP ID and its name

        LLDP ID is usually the mac address of one interface of our device. Look
        for it in netbox, then search the if_name
        """
        try:
            some_device_netif = next(
                self._mappers["interfaces"].get(mac_address=lldp_id)
            )
        except StopIteration:
            raise NetInterfaceNotFoundError(lldp_id)

        device = some_device_netif.device
        return self._get_netif_or_derivative(device.name, if_name)

    def interconnect_netbox_netif(self, netif_a, netif_b):
        """
        :returns interface_connection: wanted interface connection
        """
        props = {
            'termination_a_type': 'dcim.interface',  # because we work with physical devices only
            'termination_a_id': netif_a.id,
            'termination_b_type': 'dcim.interface',  # see above
            'termination_b_id': netif_b.id,
            'connection_status': True
        }

        netif_connection = None
        if netif_b.connected_endpoint:
            if netif_b.connected_endpoint.id == netif_a.id:
                return next(self._mappers["cables"].get(
                    netif_b.connected_endpoint.cable.id
                ))
            elif netif_a.connected_endpoint:
                self._delete_connection_to_netbox_netif(netif_b)
                # force a refresh to clear attached connections
                netif_b = next(netif_b.get())
            else:
                netif_connection = self._get_current_cable_co_of_netif(netif_b)

        if netif_a.connected_endpoint:
            netif_connection = self._get_current_cable_co_of_netif(netif_a)

        if netif_connection:
            for k, v in props.items():
                setattr(netif_connection, k, v)

            netif_connection.put()
        else:
            netif_connection = self._mappers["cables"].post(
                **props
            )

        return netif_connection

    def _delete_connection_to_netbox_netif(self, netif):
        netif_connection = self._get_current_cable_co_of_netif(netif)
        netif_connection.delete()

    def _get_current_cable_co_of_netif(self, netif):
        if not netif.connected_endpoint:
            raise ValueError(
                "No connection found for network interface {}".format(netif.id)
            )

        if netif.connected_endpoint.cable.id:
            return next(self._mappers["cables"].get(
                netif.connected_endpoint.cable.id
            ))
        else:
            raise ValueError(
                "No found cable connection for network interface {}".format(netif.id)
            )

    def _find_connection_in_netif_connections(self, netif_connections, netif):
        for c in netif_connections:
            if c.interface_a.id == netif.id or c.interface_b.id == netif.id:
                return c

        raise ValueError(
            "No connection found for network interface {}".format(netif.id)
        )

    def _get_netif_or_derivative(self, hostname, netif):
        """
        Get netif or derivative names of hostname

        netif can be a macaddr, but to be conservative, it will only work if
        only one interface has this macaddr.
        """
        interfaces = self._get_interfaces_for_device(hostname)

        if netif in interfaces:
            return interfaces[netif]

        if is_macaddr(netif):
            mac_addresses = defaultdict(list)
            for i in interfaces:
                mac_addresses[macaddr_to_int(interfaces[i].mac_address)].append(interfaces[i])

            int_netif_mac = macaddr_to_int(netif)
            if len(mac_addresses.get(int_netif_mac, 0)) == 1:
                return mac_addresses[int_netif_mac][0]
        else:
            for netif_deriv in self._get_all_derivatives_for_netif(netif):
                for k in interfaces:
                    for i in self._get_all_derivatives_for_netif(k):
                        if i == netif_deriv:
                            return interfaces[k]

        raise ValueError(
            "Interface {} not found".format(netif)
        )

    def _get_interfaces_for_device(self, hostname):
        interfaces = self.interfaces_cache.get(hostname)
        if interfaces is not None:
            return interfaces

        try:
            device = next(self._mappers["devices"].get(name=hostname))
        except StopIteration:
            raise DeviceNotFoundError(hostname)

        interfaces = {
            netif.name: netif
            for netif in self._mappers["interfaces"].get(device_id=device.id)
        }
        self.interfaces_cache[hostname] = interfaces

        return interfaces

    def _get_all_derivatives_for_netif(self, netif):
        yield netif
        yield CiscoParser.get_abrev_if(netif)
        yield JuniperParser.get_real_ifname(netif)

    def _update_discovered_from_netif_connection(self, discovered, netif_conn):
        """
        Update the discovered dict with the correct netif names

        Names received during the discovery are not always the ones really
        connected in netbox (thanks to the guessing algorithms). Add in
        the `discovered` dict with the correct ones.
        """
        netif_a = netif_conn.termination_a
        hostname_a = netif_a.device.name
        netif_b = netif_conn.termination_b
        hostname_b = netif_b.device.name

        discovered[hostname_a][netif_a.name] = (hostname_b, netif_b.name)
        discovered[hostname_b][netif_b.name] = (hostname_a, netif_a.name)
        return discovered

    @generic_netbox_error
    def _clean_undetected_intercos(self, hostname, discovered):
        for netif in self._get_interfaces_for_device(hostname).values():
            if netif.name not in discovered[hostname]:
                try:
                    self._delete_connection_to_netbox_netif(netif)
                except ValueError:
                    pass



import concurrent.futures
from concurrent.futures import ThreadPoolExecutor
import logging
import napalm
from tqdm import tqdm
import yaml

from netbox_netprod_importer.importer import DeviceImporter


logger = logging.getLogger("netbox_importer")


def parse_devices_yaml_def(devices_yaml, creds=None):
    devices = {}
    with open(devices_yaml) as devices_yaml_str:
        for hostname, props in tqdm(yaml.safe_load(devices_yaml_str).items()):
            try:
                devices[hostname] = DeviceImporter(
                    props.get("target") or hostname,
                    napalm_driver_name=props["driver"],
                    napalm_optional_args=props.get("optional_args"),
                    creds=creds,
                    discovery_protocol=props.get("discovery_protocol")
                )
            except Exception as e:
                logger.error(
                    "Cannot connect to device %s: %s", hostname, e
                )
    return devices



from enum import Enum
from .base import _AbstractVendorParser
from .cisco import NXOSParser, IOSParser
from .juniper import JunOSParser

__all__ = (
    "NXOSParser", "JunOSParser", "DeviceParsers"
)


class DeviceParsers(Enum):
    junos = JunOSParser
    nxos = NXOSParser
    nxos_ssh = NXOSParser
    ios = IOSParser


class StubParser(_AbstractVendorParser):

    def get_interfaces_lag(self, *args, **kwargs):
        return super().get_interfaces_lag(*args, **kwargs)

    def get_interface_type(self, *args, **kwargs):
        return "Other"



#!/usr/bin/env python3

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

requirements = [
    "appdirs", "cachetools", "defusedxml", "lxml", "napalm", "netboxapi",
    "simplejson", "tqdm"
]
setup_requirements = [
    "pytest-runner",
]
test_requirements = [
    "pytest", "pytest-cov", "pytest-mock",
]

setup(
    author="Scaleway",
    author_email="contact@scaleway.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    description="Microservice import into netbox devices in production",
    install_requires=requirements,
    long_description=readme,
    include_package_data=True,
    keywords="netbox_netprod_importer",
    name="netbox_netprod_importer",
    packages=find_packages(include=["netbox_netprod_importer", "netbox_netprod_importer.*"]),
    package_data={
        "netbox_netprod_importer": ["templates/**"]
    },
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://gitlab.infra.online.net/network/netbox_netprod_importer",
    version="0.2.0",
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'netbox-netprod-importer = netbox_netprod_importer.__main__:parse_args',
        ],
    }
)



import concurrent.futures
from concurrent.futures import ThreadPoolExecutor
import getpass
import json
import logging
import socket
import sys
import argparse
from netboxapi import NetboxAPI
from tqdm import tqdm

from . import __appname__, __version__
from netbox_netprod_importer.config import get_config, load_config
from netbox_netprod_importer.devices_list import parse_devices_yaml_def
from netbox_netprod_importer.push import (
    NetboxDevicePropsPusher, NetboxInterconnectionsPusher
)


logger = logging.getLogger("netbox_importer")


def parse_args():
    parser = argparse.ArgumentParser(
        description="Import into netbox network devices in production"
    )
    subcommands = parser.add_subparsers()

    sp_import = subcommands.add_parser(
        "import", aliases=["imp"], help=("import devices data")
    )
    sp_import.set_defaults(func=import_data)

    sp_interconnect = subcommands.add_parser(
        "interconnect", aliases=["interco"], help=("interconnect devices")
    )
    sp_interconnect.set_defaults(func=interconnect)

    sp_inventory = subcommands.add_parser(
        "inventory", aliases=["inv"],
        help=("inventory devices (import + interconnect)")
    )
    sp_inventory.set_defaults(func=inventory)

    for sp in (sp_import, sp_interconnect, sp_inventory):
        sp.add_argument(
            "devices", metavar="DEVICES", type=str,
            help="Yaml file containing a definition of devices to poll"
        )
        sp.add_argument(
            "-u", "--user", metavar="USER",
            help="user to use for connections to the devices",
            dest="user", type=str
        )
        sp.add_argument(
            "-p", "--password",
            help="ask for credentials for connections to the devices",
            dest="ask_password", action="store_true"
        )
        sp.add_argument(
            "-P", "--Password", metavar="PASSWORD",
            help="credentials for connections to the devices",
            dest="password", type=str
        )
        sp.add_argument(
            "-t", "--threads", metavar="THREADS",
            help="number of threads to run",
            dest="threads", default=10, type=int
        )
        sp.add_argument(
            "--overwrite",
            help="overwrite data already pushed",
            dest="overwrite", action="store_true"
        )
        sp.add_argument(
            "-v", "--verbose", metavar="LEVEL",
            help="enable debug or warning, verbose output",
            dest="verbose"
        )

    parser.add_argument(
        "--version", action="version",
        version="{} {}".format(__appname__, __version__)
    )

    arg_parser = parser
    args = arg_parser.parse_args()

    if hasattr(args, "func"):
        try:
            load_config()
        except FileNotFoundError:
            sys.exit(2)

        if args.verbose:
            numeric_level = getattr(logging, args.verbose.upper(), None)
            if not isinstance(numeric_level, int):
                raise ValueError('Invalid log level: %s' % args.verbose)
            logging.getLogger().setLevel(numeric_level)
        args.func(parsed_args=args)
    else:
        arg_parser.print_help()
        sys.exit(1)

def inventory(parsed_args):
    import_data(parsed_args)
    interconnect(parsed_args)

def import_data(parsed_args):
    creds = _get_creds(parsed_args)
    threads = parsed_args.threads

    print("Initializing importers…")
    importers = parse_devices_yaml_def(parsed_args.devices, creds)
    print()

    print("Fetching and pushing data…")
    for host, props in _multithreaded_devices_polling(
            importers=importers, threads=threads,
            overwrite=parsed_args.overwrite
    ):
        continue


def _get_creds(parsed_args):
    creds = ()
    if parsed_args.ask_password:
        creds = (parsed_args.user or getpass.getuser(), getpass.getpass())
    elif parsed_args.password:
        creds = (parsed_args.user or getpass.getuser(), parsed_args.password)

    return creds


def _multithreaded_devices_polling(importers, threads=10, overwrite=False):
    importers = importers.copy()
    netbox_api = NetboxAPI(**get_config()["netbox"])
    with ThreadPoolExecutor(max_workers=threads) as executor:
        futures = {}
        for host, importer in importers.items():
            future = executor.submit(
                _poll_and_push, netbox_api, host, importer, overwrite
            )

            futures[future] = host

        futures_with_progress = tqdm(
            concurrent.futures.as_completed(futures),
            total=len(futures)
        )
        for future in futures_with_progress:
            host = futures[future]
            try:
                yield host, future.result()
                importers.pop(host)
            except Exception as e:
                logger.error("Error when polling device %s: %s", host, e)


def _poll_and_push(netbox_api, host, importer, overwrite):
    with importer:
        props = importer.poll()
        pusher = NetboxDevicePropsPusher(
            netbox_api, host, props, overwrite=overwrite
        )
        pusher.push()

        return props


def interconnect(parsed_args):
    creds = _get_creds(parsed_args)
    threads = parsed_args.threads
    netbox_api = NetboxAPI(**get_config()["netbox"])
    remove_domains = get_config().get("remove_domains")

    interco_pusher = NetboxInterconnectionsPusher(
        netbox_api, remove_domains=remove_domains
    )

    print("Initializing importers…")
    importers = parse_devices_yaml_def(parsed_args.devices, creds)
    print()

    print("Finding neighbours and interconnecting…")
    interco_result = interco_pusher.push(
        importers=importers, threads=threads, overwrite=parsed_args.overwrite
    )
    print("{} interconnection(s) applied".format(interco_result["done"]))
    if interco_result["errors_device"]:
        logger.error(
            "Error getting neighbours on %s device(s)",
            interco_result["errors_device"]
        )
    if interco_result["errors_interco"]:
        logger.error(
            "Error pushing %s interconnection(s)",
            interco_result["errors_interco"]
        )


if __name__ == "__main__":
    parse_args()



from enum import Enum


class InterfacesRegex(Enum):
    eth100 = r".*100 Base(-T|TX)"
    eth1000 = r".*1000 Base(-T|TX)"
    sfp = r"^SFP-.?G.*"
    sfp_plus = r"^(.?WDM-)?SFP.?-.?10G.*-.*"
    qsfp_plus = r"QSFP\+"
    xfp = r"XFP"
